#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "common.h"
#include "macros.h"
#include "failures.h"
#include "hash.h"
#include "as_tree.h"
#include "symbol.h"
#include "xmalloc.h"
#include "list.h"
#include "quad.h"
#include "tree2inter.h"

/*
 * Tree2inter subsystem converts a program tree to a liner array of quads. At
 * the end of coverting we will get an array of quads. Later quads can be easily
 * mapped to the machine instructions.
 */

/* Helper data structure to process relation and logical operations.  */
struct bool_lists {
	struct slist *true_list;
	struct slist *false_list;
};
/* Helper for handlig if and while nodes. */
struct stmt_lists {
	struct slist *break_list;
	struct slist *continue_list;
};

typedef struct quad *(*represent_handler_t)(struct ast_node *, void *);

static struct quad *represent_node(struct ast_node *node, void *opaque);
static struct quad *represent_id_node(struct ast_node *node, void *opaque);
static struct quad *represent_refer_node(struct ast_node *node, void *opaque);
static struct quad *represent_op_node(struct ast_node *node, void *opaque);
static struct quad *represent_const_node(struct ast_node *node, void *opaque);
static struct quad *represent_assign_node(struct ast_node *node, void *opaque);
static struct quad *represent_not_node(struct ast_node *node, void *opaque);
static struct quad *represent_call_node(struct ast_node *node, void *opaque);
static struct quad *represent_if_node(struct ast_node *node, void *opaque);
static struct quad *represent_for_node(struct ast_node *node, void *opaque);
static struct quad *represent_while_node(struct ast_node *node, void *opaque);
static struct quad *represent_return_node(struct ast_node *node, void *opaque);
static struct quad *represent_break_node(struct ast_node *node, void *opaque);
static struct quad *represent_continue_node(struct ast_node *node, void *opaque);
static struct quad *represent_var_init_node(struct ast_node *node, void *opaque);
static struct quad *represent_array_init_node(struct ast_node *node, void *opaque);
static struct quad *represent_func_node(struct ast_node *node, void *opaque);
#ifdef _MEM_AREA
static struct quad *represent_load_mem_area_node(struct ast_node *node, void *opaque);
#endif
static struct quad *represent_load_access_node(struct ast_node *node, void *opaque);
static struct quad *represent_scope_node(struct ast_node *node, void *opaque);
static struct quad *represent_scope_end_node(struct ast_node *node, void *opaque);

/* Each type of node has a representation handler. Handler converts a node to
 * a quad or set of quads. In case of a new node type we only need to add
 * handler to this table. */
static struct {
	node_type_t node_type;
	represent_handler_t handler;
} representers[] = {
	{ NODE_TYPE_ID,		represent_id_node },
	{ NODE_TYPE_REFER,	represent_refer_node },
	{ NODE_TYPE_ACCESS,	represent_load_access_node },
#ifdef _MEM_AREA
	{ NODE_TYPE_MEM_AREA,	represent_load_mem_area_node },
#endif
	{ NODE_TYPE_MUL_OP,	represent_op_node },
	{ NODE_TYPE_SUM_OP,	represent_op_node },
	{ NODE_TYPE_REL_OP,	represent_op_node },
	{ NODE_TYPE_AND_OP,	represent_op_node },
	{ NODE_TYPE_OR_OP,	represent_op_node },
	{ NODE_TYPE_CONST,	represent_const_node },
	{ NODE_TYPE_ASSIGN,	represent_assign_node },
	{ NODE_TYPE_CALL,	represent_call_node },
	{ NODE_TYPE_IF,		represent_if_node },
	{ NODE_TYPE_FOR,	represent_for_node },
	{ NODE_TYPE_NOT,	represent_not_node },
	{ NODE_TYPE_WHILE,	represent_while_node },
	{ NODE_TYPE_RETURN,	represent_return_node },
	{ NODE_TYPE_BREAK,	represent_break_node },
	{ NODE_TYPE_CONTINUE, 	represent_continue_node },
	{ NODE_TYPE_FUNCTION, 	represent_func_node },
	{ NODE_TYPE_VAR_INIT, 	represent_var_init_node },
	{ NODE_TYPE_ARG_INIT, 	represent_var_init_node },
	{ NODE_TYPE_ARRAY_INIT,	represent_array_init_node },
	{ NODE_TYPE_SCOPE,	represent_scope_node },
	{ NODE_TYPE_SCOPE_END, 	represent_scope_end_node },
	{ NODE_TYPE_UNKNOWN,	NULL }
};

static struct quad	**quads;	/* result array of quads */
static unsigned int	quads_used;	/* number element in use  */
static unsigned int	quads_size;	/* quads array size  */

static int		label_counter;	/* counter for generating labels */
static unsigned int	tmp_counter;	/* counter for temporary variables */
static struct slist	*ret_list;	/* list of function return values which
					   should be backpatched */

static struct quad_id *mk_temp(data_type_t data_type);

static int gen_label()
{
	label_counter++;
	return label_counter;
}

/* Backpatch helper, we backpatch goto quads when we create appropriate label. */
static void backpatch(struct slist *list, int patch)
{
	struct slist *iter;
	int *data;

	return_if_fail(list != NULL);

	for (iter = list; iter != NULL; iter = iter->next) {
		data = (int *)iter->data;
		*(data) = patch;
	}
}

/* Add quad to the result array. */
static void put_quad(struct quad *q)
{
	if (quads_used == quads_size) {
		quads_size += 16;
		quads = xrealloc(quads, sizeof(struct quad *)*quads_size);
	}
	quads[quads_used++] = q;
}

/* Helper for creating a temporary quad argument. */
static struct quad_id *mk_temp(data_type_t data_type)
{
	struct quad_id *arg;
	char *str;

	asprintf(&str,"t%d", tmp_counter++);

	arg = quad_id(data_type, str);
	arg->temp = TRUE;

	xfree(str);

	return arg;
}
/* Create a temporary variable, explore an id node and build an assignment quad. */
static struct quad *represent_id_node(struct ast_node *node, void *opaque)
{
	struct ast_node_id *id_node;
	struct quad_id *rvalue;
	struct quad_id *lvalue;
	struct quad_assign *assign;

	return_val_if_fail(node != NULL, NULL);

	id_node = (struct ast_node_id *)node;

	rvalue = quad_id(id_node->decl_type, id_node->name);
	lvalue = mk_temp(id_node->decl_type);

	assign = quad_assign(QUAD(lvalue), QUAD(rvalue));
	put_quad(QUAD(assign));

	return assign->lvalue;
}

/* Create a temporary variable, explore a reference node and build an assignment
 * quad. */
static struct quad *represent_refer_node(struct ast_node *node, void *opaque)
{
	struct ast_node_refer *refer_node;
	struct quad_refer *rvalue;
	struct quad_id *lvalue;
	struct quad_assign *assign;

	return_val_if_fail(node != NULL, NULL);

	refer_node = (struct ast_node_refer *)node;

	rvalue = quad_refer(refer_node->decl_type, refer_node->name);
	lvalue = mk_temp(refer_node->decl_type);

	assign = quad_assign(QUAD(lvalue), QUAD(rvalue));
	put_quad(QUAD(assign));

	return assign->lvalue;
}

/* Create a temporary variable, a constant argument and build an assignment node. */
static struct quad *represent_const_node(struct ast_node *node, void *opaque)
{
	struct ast_node_const *_const;
	struct quad_const *rvalue;
	struct quad_id *lvalue;
	struct quad_assign *assign;

	return_val_if_fail(node != NULL, NULL);

	_const = (struct ast_node_const *)node;

	switch (_const->data_type) {
	case DATA_TYPE_INTEGER:
		rvalue = quad_const(_const->data_type, (void *)&_const->integer);
		break;
	case DATA_TYPE_REAL:
		rvalue = quad_const(_const->data_type, (void *)&_const->real);
		break;
	case DATA_TYPE_CHAR:
		rvalue = quad_const(_const->data_type, (void *)&_const->ch);
		break;
	case DATA_TYPE_STRING:
		rvalue = quad_const(_const->data_type, (void *)_const->str);
		break;
	default:
		xerror(1, "unknown constant data type");
	}

	lvalue = mk_temp(_const->data_type);

	assign = quad_assign(QUAD(lvalue), QUAD(rvalue));
	put_quad(QUAD(assign));

	return assign->lvalue;
}

/* Helper for creating an assignment quad with an integer constant */
static struct quad *represent_integer(int num)
{
	struct quad_const *rvalue;
	struct quad_id *lvalue;
	struct quad_assign *assign;

	rvalue = quad_const(DATA_TYPE_INTEGER, &num);
	lvalue = mk_temp(DATA_TYPE_INTEGER);

	assign = quad_assign(QUAD(lvalue), QUAD(rvalue));
	put_quad(QUAD(assign));

	return assign->lvalue;
}

/* Represent variable initialization node. */
static struct quad *represent_var_init_node(struct ast_node *node, void *opaque)
{
	struct ast_node_var_init *var_init;
	struct quad_assign *assign_quad;
	struct quad *lvalue, *rvalue;
	struct symbol *s;

	return_val_if_fail(node != NULL, NULL);

	var_init = (struct ast_node_var_init *)node;

	s = AST_INIT(var_init)->symbol;

	lvalue = (struct quad *)quad_id(s->decl_type, s->name);

	rvalue = represent_node(var_init->expr, NULL);
	quad_ref(rvalue);

	assign_quad = quad_assign(lvalue, rvalue);
	put_quad(QUAD(assign_quad));

	return NULL;
}

/* Array initialization node converts to a set of store access quads. */
static struct quad *represent_array_init_node(struct ast_node *node, void *opaque)
{
	struct ast_node_array_init *array_init;
	struct quad *idx, *rvalue;
	struct quad_store_access *store;
	struct symbol *s;
	unsigned int i;

	return_val_if_fail(node != NULL, NULL);

	array_init = (struct ast_node_array_init *)node;

	s = AST_INIT(array_init)->symbol;

	for (i = 0; i < array_init->ninitvals; i++) {
		struct array_idx init = array_init->initvals[i];

		rvalue = represent_node(init.expr, NULL);
		quad_ref(rvalue);

		idx = represent_integer(init.nidx);
		quad_ref(idx);

		store = quad_store_access(s->name, idx, rvalue);
		put_quad(QUAD(store));
	}

	return NULL;
}

/* Represent a scope node. */
static struct quad *represent_scope_node(struct ast_node *node, void *opaque)
{
	struct ast_node_scope *scope;
	struct quad_scope *scope_quad;
	struct quad *arg;

	return_val_if_fail(node != NULL, NULL);

	scope = (struct ast_node_scope *)node;

	symbol_table_ref(scope->table);
	scope_quad = quad_scope(scope->table);
	put_quad(QUAD(scope_quad));

	arg = represent_node(scope->stmt, opaque);

	return arg;
}

/* Scope end quad just shows that we leave current scope. */
static struct quad *represent_scope_end_node(struct ast_node *node, void *opaque)
{
	struct quad_scope_end *scope_end;

	return_val_if_fail(node != NULL, NULL);

	scope_end = quad_scope_end();
	put_quad(QUAD(scope_end));

	return NULL;
}

/* Create array indexes, we use an array dimension to create quads which
 * calculate the liner index number. */
static struct quad *represent_access_idx(struct ast_node_access *access)
{
	struct quad_op *add_op;
	struct quad_id *idx;
	struct quad *right, *left, *result;
	unsigned int i, linear_idx;

	if (access->nidxs == 1)
		return represent_node(access->idxs[0], NULL);

	left = represent_node(access->idxs[0], NULL);
	quad_ref(left);

	linear_idx = 1;
	for (i = 1; i < access->nidxs; i++)
		linear_idx *= access->dims[i];

	right = represent_integer(linear_idx);
	quad_ref(right);
	idx = mk_temp(DATA_TYPE_INTEGER);

	result = QUAD(idx);

	add_op = quad_op(OPCODE_MUL, result, left, right);
	put_quad(QUAD(add_op));

	for (i = 1; i < access->nidxs - 1; i++) {
		struct quad_op *mul_op;
		unsigned int j;

		left = represent_node(access->idxs[i], QUAD(mul_op));
		quad_ref(left);

		linear_idx = 1;
		for (j = i+1; j < access->nidxs; j++)
			linear_idx *= access->dims[j];

		right = represent_integer(linear_idx);
		quad_ref(right);

		result = QUAD(mk_temp(DATA_TYPE_INTEGER));

		mul_op = quad_op(OPCODE_MUL, result, left, right);
		put_quad(QUAD(mul_op));

		left = QUAD(idx);
		quad_ref(left);

		right = mul_op->result;
		quad_ref(right);

		idx = mk_temp(DATA_TYPE_INTEGER);
		result = QUAD(idx);

		add_op = quad_op(OPCODE_ADD, result, left, right);
		put_quad(QUAD(add_op));
	}

	left = QUAD(idx);
	quad_ref(left);

	right = represent_node(access->idxs[access->nidxs-1], QUAD(add_op));
	quad_ref(right);

	idx = mk_temp(DATA_TYPE_INTEGER);
	result = QUAD(idx);

	add_op = quad_op(OPCODE_ADD, result, left, right);
	put_quad(QUAD(add_op));

	return QUAD(idx);
}

static struct quad *represent_load_access_node(struct ast_node *node, void *opaque)
{
	struct ast_node_access *access;
	struct quad_load_access *load;
	struct quad *idx, *lvalue;

	return_val_if_fail(node != NULL, NULL);

	access = (struct ast_node_access *)node;

	idx = represent_access_idx(access);
	quad_ref(idx);

	lvalue = QUAD(mk_temp(access->decl_type));

	load = quad_load_access(access->name, lvalue, idx);
	put_quad(QUAD(load));

	return load->lvalue;
}

/* Left child of an assignment is an access node, we create store access quad. */
static struct quad *represent_store_access_node(struct ast_node *left,
						struct quad *rvalue, boolean_t leave)
{
	struct ast_node_access *access;
	struct quad_store_access *store;
	struct quad *idx;

	return_val_if_fail(left != NULL, NULL);
	return_val_if_fail(rvalue != NULL, NULL);

	access = (struct ast_node_access *)left;

	idx = represent_access_idx(access);
	quad_ref(idx);

	quad_ref(rvalue);
	store = quad_store_access(access->name, idx, rvalue);
	store->leave = leave;
	put_quad(QUAD(store));

	return store->rvalue;
}

#ifdef _MEM_AREA
static struct quad *represent_load_mem_area_node(struct ast_node *node, void *opaque)
{
	struct ast_node_mem_area *mem_area;
	struct quad_load_mem_area *load;
	data_type_t data_type;

	return_val_if_fail(node != NULL, NULL);

	mem_area = (struct ast_node_mem_area *)node;

	switch (mem_area->mem_data_type) {
	case 'B':
		data_type = DATA_TYPE_CHAR;
		break;
	case 'S':
	case 'L':
		data_type = DATA_TYPE_INTEGER;
		break;
	default:
		break;
	}
	load = quad_load_mem_area();

	load->area = mem_area->area;
	load->mem_data_type = mem_area->mem_data_type;
	load->offset = mem_area->offset;
	load->nbits = mem_area->nbits;
	load->is_unsign = mem_area->is_unsign;

	if (mem_area->idx != NULL) {
		load->idx = represent_node(mem_area->idx, QUAD(load));
		quad_ref(load->idx);
	}

	load->lvalue = QUAD(mk_temp(data_type));

	put_quad(QUAD(load));

	return load->lvalue;
}

/* Left child of an assignment is a memory area node. */
static struct quad *represent_store_mem_area_node(struct ast_node *left,
						  struct quad *rvalue, boolean_t leave)
{
	struct ast_node_mem_area *mem_area;
	struct quad_store_mem_area *store;

	return_val_if_fail(left != NULL, NULL);
	return_val_if_fail(rvalue != NULL, NULL);

	mem_area = (struct ast_node_mem_area *)left;

	store = quad_store_mem_area();
	store->area = mem_area->area;
	store->mem_data_type = mem_area->mem_data_type;
	store->offset = mem_area->offset;
	store->nbits = mem_area->nbits;
	store->is_unsign = mem_area->is_unsign;

	quad_ref(rvalue);
	store->rvalue = rvalue;
	store->leave = leave;

	if (mem_area->idx != NULL) {
		store->idx = represent_node(mem_area->idx, QUAD(store));
		quad_ref(store->idx);
	}

	put_quad(QUAD(store));

	return store->rvalue;
}
#endif

/* Left child of an assignment is an id or a reference node node. */
static struct quad *represent_common_assignment(struct ast_node *left,
						struct quad *rvalue, boolean_t leave)
{
	struct ast_node_id *id_node;
	struct ast_node_refer *refer_node;
	struct quad_assign *assign_quad;
	struct quad *lvalue;

	return_val_if_fail(left != NULL, NULL);
	return_val_if_fail(rvalue != NULL, NULL);

	switch (left->type) {
	case NODE_TYPE_ID:
		id_node = (struct ast_node_id *)left;
		lvalue = (struct quad *)quad_id(id_node->decl_type, id_node->name);
		break;
	case NODE_TYPE_REFER:
		refer_node = (struct ast_node_refer *)left;
		lvalue = (struct quad *)quad_refer(refer_node->decl_type, refer_node->name);
		break;
	default:
		xerror(1, "unexpected lvalue node type");
	}

	quad_ref(rvalue);
	assign_quad = quad_assign(lvalue, rvalue);
	assign_quad->leave = leave;
	put_quad(QUAD(assign_quad));

	return assign_quad->rvalue;
}

/* Left child of an assign node can be three different types, according to
 * these types we call subhandlers. */
static struct quad *represent_assign_node(struct ast_node *node, void *opaque)
{
	struct ast_node_assign *assign;
	struct quad *rvalue;
	unsigned int i;
	boolean_t leave;

	return_val_if_fail(node != NULL, NULL);

	assign = (struct ast_node_assign *)node;

	rvalue = represent_node(assign->right, NULL);
	
	for (i = 0; i < assign->nleft; i++) {
		leave = (i != assign->nleft-1) ? TRUE : FALSE;
		switch (assign->left[i]->type) {
		case NODE_TYPE_ACCESS:
			represent_store_access_node(assign->left[i], rvalue, leave);
			break;
#ifdef _MEM_AREA
		case NODE_TYPE_MEM_AREA:
			represent_store_mem_area_node(assign->left[i], rvalue, leave);
			break;
#endif
		case NODE_TYPE_ID:
		case NODE_TYPE_REFER:
			represent_common_assignment(assign->left[i], rvalue, leave);
			break;
		default:
			xerror(1, "unexpected node type");
		}
	}

	return NULL;
}

/* Represent return value. Create goto quad, we will patch its value when create
 * a label at the end of a function. */
static struct quad *represent_return_node(struct ast_node *node, void *opaque)
{
	struct ast_node_return *ret;
	struct quad_goto *goto_quad;;
	struct quad *ret_value;
	struct slist *need_patch;

	return_val_if_fail(node != NULL, NULL);

	ret = (struct ast_node_return *)node;

	ret_value = NULL;

	if (ret->ret_val_node != NULL)
		ret_value = represent_node(ret->ret_val_node, NULL);

	goto_quad = quad_goto(-1);

	need_patch = slist_new();
	need_patch->data = &(goto_quad->ln);
	ret_list = slist_add_tail(ret_list, need_patch);

	put_quad(QUAD(goto_quad));

	return QUAD(ret_value);
}

/* Represent function node. First of all explore a function tree, after that
 * if return values were found we need to backpatch them. */
static struct quad *represent_func_node(struct ast_node *node, void *opaque)
{
	struct ast_node_func *func;
	struct quad_func *func_quad;
	struct quad_label *label_quad;
	struct quad_func_end *func_end_quad;
	int ln;

	return_val_if_fail(node != NULL, NULL);

	func = (struct ast_node_func *)node;

	func_quad = quad_func(func->name, func->ret_type);
	put_quad(QUAD(func_quad));

	represent_node(func->func_tree, NULL);

	if (ret_list != NULL) {
		ln = gen_label();
		label_quad = quad_label(ln);

		backpatch(ret_list, ln);
		put_quad(QUAD(label_quad));
		slist_free_all(&ret_list, NULL);
	}

	func_end_quad = quad_func_end();
	put_quad(QUAD(func_end_quad));

	return NULL;
}

/*
 * In our language user can use logic and relation operations in assignment
 * expressions, as the result we should assign 0 -- in case of false or 1 --
 * in case of true.
 *
 * We create quads in this order:
 *
 * true_label:
 * 	t = 1
 * 	goto next_label
 * false_label:
 * 	t = 0
 * next_label:
 *
 */

static struct quad *assign_bool_expr(struct bool_lists *bool_expr)
{
	struct quad_label *label_true, *label_false, *label_next;
        struct quad_assign *assign;
        struct quad_goto *goto_next;
        struct quad *arg;
        struct quad *lvalue, *rvalue;
        int ln;
        int zero;

        /* Create true label. */
        ln = gen_label();
        label_true = quad_label(ln);
        put_quad(QUAD(label_true));

        backpatch(bool_expr->true_list, label_true->ln);
        slist_free_all(&bool_expr->true_list, NULL);
        /* Create true assignment quad. */
        arg = represent_integer(1);
        /* Goto the next label. */
        goto_next = quad_goto(-1);
        put_quad(QUAD(goto_next));

        ln = gen_label();
        label_false = quad_label(ln);
        put_quad(QUAD(label_false));
        /* Create false assignment quad. */
        zero = 0;
        rvalue = (struct quad *)quad_const(DATA_TYPE_INTEGER, &zero);

        lvalue = arg;
        quad_ref(QUAD(arg));

        assign = quad_assign(lvalue, rvalue);
        put_quad(QUAD(assign));

        backpatch(bool_expr->false_list, label_false->ln);
        slist_free_all(&bool_expr->false_list, NULL);

        ln = gen_label();
        label_next = quad_label(ln);
        goto_next->ln = label_next->ln;

        put_quad(QUAD(label_next));

        return QUAD(arg);
}

/* Represent not node. We just switch lists. Also handle case when using not
 * node in an assignment expression. */
static struct quad *represent_not_node(struct ast_node *node, void *opaque)
{
	struct ast_node_not *not;
	struct bool_lists *bool_expr, bool_not;
	struct slist *tmp_list;
	struct quad *arg;

	return_val_if_fail(node != NULL, NULL);

	not = (struct ast_node_not *)node;

	memset(&bool_not, 0, sizeof(struct bool_lists));

	if (opaque == NULL)
		bool_expr = &bool_not;
	else
		bool_expr = (struct bool_lists *)opaque;

	represent_node(not->expr, QUAD(bool_expr));

	/* Change true to false, false to true */
	tmp_list = bool_expr->true_list;
	bool_expr->true_list = bool_expr->false_list;
	bool_expr->false_list = tmp_list;

	arg = NULL;
	/* Case of an arithmetic operation */
	if (opaque == NULL)
		arg = assign_bool_expr(bool_expr);

	return arg;
}

/* Represent arithmetic operations. We just create operation quad. */
static struct quad *represent_arith_op_node(struct ast_node *node, void *opaque)
{
	struct ast_node_op *op;
	struct quad_op *op_quad;
	struct quad *result, *left, *right;
	data_type_t data_type;

	return_val_if_fail(node != NULL, NULL);

	op = (struct ast_node_op *)node;

	left = represent_node(op->left, NULL);
	quad_ref(left);

	right = represent_node(op->right, NULL);
	quad_ref(right);

	data_type = QUAD_ARG(left)->data_type;
	result = QUAD(mk_temp(data_type));

	op_quad = quad_op(op->opcode, result, left, right);
	put_quad(QUAD(op_quad));

	return QUAD(result);
}

/* Represent call node. First of all we put an empty constant quad for return
 * value, then map each argument to param quad using function symbol and node.
 * Also we check cases when we don't need return id_quad as return value.*/
static struct quad *represent_call_node(struct ast_node *node, void *opaque)
{
	struct ast_node_func_call *call;
	struct quad_call *call_quad;
	struct quad_param *param_quad;
	struct quad_id *ret_val;
	struct quad *param;

	struct symbol_table *global;
	struct symbol *s;
	ret_t ret;
	int i;

	return_val_if_fail(node != NULL, NULL);

	call = (struct ast_node_func_call *)node;

	global = symbol_table_get_global_table();
	ret = hash_table_lookup(global->scope, call->name, (void **)&s);
	if (ret != ret_ok)
		xerror(1, "can't detect a function symbol %s", call->name);

	/* Additional param quad to prepare place for the return value. */
	if (call->ret_type != DATA_TYPE_VOID) {
		struct quad_const *_const;
		char *str = "";
		char ch = '0';
		int integer = 0;
		float real = 0.0;

		switch (call->ret_type) {
		case DATA_TYPE_INTEGER:
			_const = quad_const(DATA_TYPE_INTEGER, &integer);
			break;
		case DATA_TYPE_CHAR:
			_const = quad_const(DATA_TYPE_CHAR, &ch);
			break;
		case DATA_TYPE_REAL:
			_const = quad_const(DATA_TYPE_REAL, &real);
			break;
		case DATA_TYPE_STRING:
			_const = quad_const(DATA_TYPE_STRING, str);
			break;
		default:
			break;
		}
		param_quad = quad_param(QUAD(_const), FALSE);
		put_quad(QUAD(param_quad));
	}

	ret_val = NULL;
	/* Call quad can have NULL `ret_val' field in case of the void return
	 * value and in case if nobody is going to take the return value after
	 * calling a function (user just calls a function and doesn't take a
	 * return value). */
	if (AST_NODE(call)->parent != NULL) {
		struct ast_node *parent = AST_NODE(call)->parent;
		if (parent->type != NODE_TYPE_STMT &&
				call->ret_type != DATA_TYPE_VOID) {
			ret_val = mk_temp(call->ret_type);
		}
	}

	call_quad = quad_call(call->name, QUAD(ret_val), call->ret_type);
	/* Represent all arguments as param quads. Put them in the reverse order. */
	for (i = call->nargs-1; i >= 0; i--) {
		boolean_t is_ref = FALSE;
		data_type_t data_type;
		/* Call can have more arguments than a function prototype.
		 * Also there is a special handling for reference arguments. */
		if (i < SYMBOL_FUNC(s)->nargs && SYMBOL_FUNC(s)->args[i]->is_ref) {
			struct ast_node_id *id_node;
			struct ast_node_refer *refer_node;

			is_ref = TRUE;
			switch (call->args[i]->type) {
			case NODE_TYPE_ID:
				id_node = (struct ast_node_id *)call->args[i];
				param = (struct quad *)quad_id(id_node->decl_type,
									id_node->name);
				data_type = id_node->decl_type;
				break;
			case NODE_TYPE_REFER:
				refer_node = (struct ast_node_refer *)call->args[i];
				param = (struct quad *)quad_refer(refer_node->decl_type,
									refer_node->name);
				data_type = refer_node->decl_type;
				break;
			default:
				xerror(1, "unexpected reference type");
			}
		} else {
			param = represent_node(call->args[i], NULL);
			quad_ref(param);
			data_type = QUAD_ARG(param)->data_type;
		}

		quad_call_add_arg(call_quad, data_type, is_ref);
		param_quad = quad_param(param, is_ref);
		put_quad(QUAD(param_quad));
	}
	

	put_quad(QUAD(call_quad));
	return QUAD(ret_val);
}

/* Represent continue node, it is simple goto quad, we will patch continue goto
 * quad when create a next step cycle label. */
static struct quad *represent_continue_node(struct ast_node *node, void *opaque)
{
	struct stmt_lists *parent_quad;
	struct quad_goto *goto_begin;
	struct slist *need_patch;

	return_val_if_fail(node != NULL, NULL);
	return_val_if_fail(opaque != NULL, NULL);

	parent_quad = (struct stmt_lists *)opaque;

	goto_begin = quad_goto(-1);
	put_quad(QUAD(goto_begin));

	need_patch = slist_new();
	need_patch->data = &(goto_begin->ln);

	parent_quad->continue_list = slist_add_tail(parent_quad->continue_list, need_patch);

	return NULL;
}

/* Represent break node, it is simple goto quad, we will patch break goto quad
 * when create an exit cycle label. */
static struct quad *represent_break_node(struct ast_node *node, void *opaque)
{
	struct stmt_lists *parent_quad;
	struct quad_goto *goto_next;
	struct slist *need_patch;

	return_val_if_fail(node != NULL, NULL);
	return_val_if_fail(opaque != NULL, NULL);

	parent_quad = (struct stmt_lists *)opaque;

	goto_next = quad_goto(-1);
	put_quad(QUAD(goto_next));

	need_patch = slist_new();
	need_patch->data = &(goto_next->ln);

	parent_quad->break_list = slist_add_tail(parent_quad->break_list, need_patch);

	return NULL;
}

/*
 *  Represent relation operation, it consist of 2 quads.
 * 	if `left' op_type `right' goto true_label
 * 	goto false_label
 *
 *	If we don't have a parent, it means that we use a relation operation in
 *	an assignment expression. This case is handled in assign_bool_expr
 *      function.
 *
 */
static struct quad *represent_rel_op_node(struct ast_node *node, void *opaque)
{
	struct ast_node_op *rel;
	struct quad_if *if_quad;
	struct quad_goto *goto_false;
	struct bool_lists *bool_expr, bool_rel;
	struct quad *q, *left, *right;
	struct slist *need_patch;

	return_val_if_fail(node != NULL, NULL);

	rel = (struct ast_node_op *)node;
	memset(&bool_rel, 0, sizeof(struct bool_lists));

	if (opaque == NULL)
		bool_expr = &bool_rel;
	else
		bool_expr = (struct bool_lists *)opaque;
	/* Represent left node */
	left = represent_node(rel->left, NULL);
	quad_ref(left);
	/* Represent right node */
	right = represent_node(rel->right, NULL);
	quad_ref(right);
	/* Create the first quad. */
	if_quad = quad_if(rel->opcode, left, right, -1);
	put_quad(QUAD(if_quad));
	/* Create lists for true and false jumps, now we don't have these labels,
	 * these labels will be created later and in that time we will patch lists
	 * values. */
	need_patch = slist_new();
	need_patch->data = &(if_quad->ln);
	bool_expr->true_list = slist_add_tail(bool_expr->true_list, need_patch);

	goto_false = quad_goto(-1);
	need_patch = slist_new();
	need_patch->data = &(goto_false->ln);
	bool_expr->false_list = slist_add_tail(bool_expr->false_list, need_patch);

	put_quad(QUAD(goto_false));

	q = NULL;
	/* Handle an assignment expression case. */
	if (opaque == NULL)
		q = assign_bool_expr(bool_expr);

	return q;
}

/*
 *  Here we implement simple scheme:
 *
 *      B -> B1 || M B2 
 *      (where B1 -- bool_left, B2 -- bool_right, B -- parent bool_expr list,
 *      M -- label number)
 *
 *      1) backpatch(B1.false_list, M.instr)
 *              We only can patch false list of left node.
 *
 *      2) B.true_list = merge(B1.true_list, B2.true_list)
 *              False lists of left and right nodes should be merged and given
 *              to our parent.
 *	
 *	3) B.false_list = B2.false_list
 *		False list we give to our parent.
 *
 *      If we don't have a parent, it means that we use logic operation in an
 *      assignment expression. This case is handled in assign_bool_expr
 *      function.
 */
static struct quad *process_or_op(struct ast_node *node, void *opaque)
{
	struct ast_node_op *logic;
	struct bool_lists bool_left, bool_right, *bool_expr, bool_or;
	struct quad_label *label_false;
	struct quad *q;
	int ln;

	return_val_if_fail(node != NULL, NULL);

	logic = (struct ast_node_op *)node;

	memset(&bool_left, 0, sizeof(struct bool_lists));
	memset(&bool_right, 0, sizeof(struct bool_lists));
	memset(&bool_or, 0, sizeof(struct bool_lists));
	/* Represent left node, in our scheme it is B1. */
	represent_node(logic->left, &bool_left);
	/* Create true label, in scheme it is M. */
	ln = gen_label();
	label_false = quad_label(ln);
	put_quad(QUAD(label_false));
	/* 1)  */
	backpatch(bool_left.false_list, label_false->ln);
	slist_free_all(&bool_left.false_list, NULL);
	/* Represent left node, in our scheme it is B2. */
	represent_node(logic->right, &bool_right);

	if (opaque == NULL) {
		bool_expr = &bool_or;
		/* 2) */
		bool_expr->true_list = slist_add_tail(bool_left.true_list, bool_right.true_list);
		/* 3) */
		bool_expr->false_list = bool_right.false_list;
		/* Handle an assignment expression case. */
		q = assign_bool_expr(bool_expr);
	} else {
		bool_expr = (struct bool_lists *)opaque;
		/* 2) */
		bool_expr->true_list = slist_add_tail(bool_left.true_list, bool_right.true_list);
		/* 3) */
		bool_expr->false_list = bool_right.false_list;
	}

	return q;
}

/*
 *  Here we implement simple scheme:
 *
 *	B -> B1 && M B2 
 *	(where B1 -- bool_left, B2 -- bool_right, B -- parent bool_expr list,
 *	M -- label number)
 *
 *	1) backpatch(B1.true_list, M.instr)
 *		We only can patch true list of left node.
 *
 *	2) B.true_list = B2.true_list
 *		True list we give to our parent.
 *
 *	3) B.false_list = merge(B1.false_list, B2.false_list)
 *		False lists of left and right nodes should be merged and given
 *		to our parent.
 *
 *	If we don't have a parent, it means that we use logic operation in an
 *	assignment expression. This case is handled in assign_bool_expr
 *	function.
 */
static struct quad *process_and_op(struct ast_node *node, void *opaque)
{
	struct ast_node_op *logic;
	struct bool_lists bool_left, bool_right, *bool_expr, bool_and;
	struct quad_label *label_true;
	struct quad *q;
	int ln;

	return_val_if_fail(node != NULL, NULL);

	logic = (struct ast_node_op *)node;

	memset(&bool_left, 0, sizeof(struct bool_lists));
	memset(&bool_right, 0, sizeof(struct bool_lists));
	memset(&bool_and, 0, sizeof(struct bool_lists));
	
	/* Represent left node, in our scheme it is B1. */
	represent_node(logic->left, &bool_left);
	/* Create true label, in scheme it is M. */
	ln = gen_label();
	label_true = quad_label(ln);
	put_quad(QUAD(label_true));
	/* 1) */
	backpatch(bool_left.true_list, label_true->ln);
	slist_free_all(&bool_left.true_list, NULL);
	/* Represent right node, in our scheme it is B2. */
	represent_node(logic->right, &bool_right);

	q = NULL;
	/* Does a parent send us his own bool_expr? */
	if (opaque == NULL) {
		bool_expr = &bool_and;
		/* 2) */
		bool_expr->true_list = bool_right.true_list;
		/* 3) */
		bool_expr->false_list = slist_add_tail(bool_left.false_list, bool_right.false_list);
		/* Handle an assignment expression case. */
		q = assign_bool_expr(bool_expr);
	} else {
		bool_expr = (struct bool_lists *)opaque;
		/* 2) */
		bool_expr->true_list = bool_right.true_list;
		/* 3) */
		bool_expr->false_list = slist_add_tail(bool_left.false_list, bool_right.false_list);
	}

	return q;
}

/* Represent `or' and `and' operation nodes. */
static struct quad *represent_logic_op_node(struct ast_node *node, void *opaque)
{
	struct ast_node_op *logic;
	struct quad *q;

	return_val_if_fail(node != NULL, NULL);

	logic = (struct ast_node_op *)node;
	q = NULL;

	switch (logic->opcode) {
	case OPCODE_OR:
		q = process_or_op(node, opaque);
		break;
	case OPCODE_AND:
		q = process_and_op(node, opaque);
		break;
	default:
		break;
	}

	return q;
}

/* Represent operation node. For special types of opertaion we call different
 * subhandlers.  */
static struct quad *represent_op_node(struct ast_node *node, void *opaque)
{
	struct ast_node_op *op;
	struct quad *q;

	return_val_if_fail(node != NULL, NULL);

	op = (struct ast_node_op *)node;
	q = NULL;

	switch (op->opcode) {
	case OPCODE_OR:
	case OPCODE_AND:
		q = represent_logic_op_node(node, opaque);
		break;
	case OPCODE_ADD:
	case OPCODE_SUB:
	case OPCODE_MOD:
	case OPCODE_MUL:
	case OPCODE_DIV:
		q = represent_arith_op_node(node, opaque);
		break;
	case OPCODE_EQ:
	case OPCODE_NE:
	case OPCODE_GE:
	case OPCODE_GT:
	case OPCODE_LE:
	case OPCODE_LT:
		q = represent_rel_op_node(node, opaque);
		break;
	default:
		xerror(1, "unknown opcode type");
	}

	return q;
}

/*
 * Represen while node, below there is a scheme.
 *
 * :label_begin
 * 	while_expr
 * :label_true
 * 	while_stmt
 *  	goto label_begin
 * :label_next
 *
 */
static struct quad *represent_while_node(struct ast_node *node, void *opaque)
{
	struct ast_node_while *while_node;

	struct quad_label *label_begin, *label_true, *label_next;
	struct quad_goto *goto_begin;

	struct bool_lists while_bool;
	struct stmt_lists while_stmt;
	int ln;

	return_val_if_fail(node != NULL, NULL);

	while_node = (struct ast_node_while *)node;

	memset(&while_bool, 0, sizeof(struct bool_lists));
	memset(&while_stmt, 0, sizeof(struct stmt_lists));

	/* First of all create the begin label. */
	ln = gen_label();
	label_begin = quad_label(ln);
	put_quad(QUAD(label_begin));

	/* Explore a conditional expression. */
	represent_node(while_node->expr, &while_bool);
	/* Create a true label. */
	ln = gen_label();
	label_true = quad_label(ln);
	put_quad(QUAD(label_true));
	/* Backpatch values which need to jump to the true label. */
	backpatch(while_bool.true_list, label_true->ln);
	slist_free_all(&while_bool.true_list, NULL);
	/* Explore a cycle body. */
	represent_node(while_node->stmt, &while_stmt);
	/* If we found continue nodes during a body exploration -- backpatch values
	 * with a begin label. */
	if (while_stmt.continue_list != NULL) {
		backpatch(while_stmt.continue_list, label_begin->ln);
		slist_free_all(&while_stmt.continue_list, NULL);
	}
	/* Go to the beginning of the cycle. */
	ln = label_begin->ln;
	goto_begin = quad_goto(ln);
	put_quad(QUAD(goto_begin));
	/* Create exit label. */
	ln = gen_label();
	label_next = quad_label(ln);
	put_quad(QUAD(label_next));
	/* We need to make 2 backpatches -- for false expressions and
	 * for break nodes. */
	backpatch(while_bool.false_list, label_next->ln);
	slist_free_all(&while_bool.false_list, NULL);

	if (while_stmt.break_list != NULL) {
		backpatch(while_stmt.break_list, label_next->ln);
		slist_free_all(&while_stmt.break_list, NULL);
	}

	return NULL;
}

/* Represent if node, below there is the scheme of result quads.
 * 	if_expr
 * :label_true
 * 	if_stmt
 * 	goto label_next
 * :label_false
 * 	else_stmt
 * :label_next
 *
 */
static struct quad *represent_if_node(struct ast_node *node, void *opaque)
{
	struct ast_node_if *if_node;
	struct quad_label *label_true, *label_false, *label_next;
	struct quad_goto *goto_next;
	struct bool_lists if_bool;
	int ln;

	return_val_if_fail(node != NULL, NULL);

	if_node = (struct ast_node_if *)node;

	memset(&if_bool, 0, sizeof(struct bool_lists));
	/* Explore a condition expression. */
	represent_node(if_node->expr, &if_bool);
	/* Create a true label. */
	ln = gen_label();
	label_true = quad_label(ln);
	put_quad(QUAD(label_true));

	backpatch(if_bool.true_list, label_true->ln);
	slist_free_all(&if_bool.true_list, NULL);
	/* Explorer the `if' body, opaque can contain a stmt_lists. */
	represent_node(if_node->stmt, opaque);

	if (if_node->_else != NULL) {
		/* If an else statement is presented -- goto the next label. */
		goto_next = quad_goto(-1);
		put_quad(QUAD(goto_next));
		/* Create a false label. */
		ln = gen_label();
		label_false = quad_label(ln);
		put_quad(QUAD(label_false));
		/* Explore an else statement. */
		represent_node(if_node->_else, opaque);
	}

	ln = gen_label();
	label_next = quad_label(ln);
	put_quad(QUAD(label_next));
	/* Two different cases of jumping to the false label. */
	if (if_node->_else != NULL) {
		goto_next->ln = label_next->ln;
		backpatch(if_bool.false_list, label_false->ln);
	} else {
		backpatch(if_bool.false_list, label_next->ln);
	}

	slist_free_all(&if_bool.false_list, NULL);

	return NULL;
}

/*
 * Represent for node, below there is the scheme of result quads. If some
 * optional expressions are absent, it will look a litle bit different.
 *
 * 	opt_expr1
 * :label_begin
 * 	opt_expr2
 * :label_true
 *	for_stmt
 * :label_next 
 *	opt_expr3
 *	goto label_begin
 * :label_false
 *
 */

static struct quad *represent_for_node(struct ast_node *node, void *opaque)
{
	struct ast_node_for *for_node;

	struct quad_label *label_begin, *label_true, *label_next, *label_false;
	struct quad_goto *goto_begin;

	struct bool_lists for_bool;
	struct stmt_lists for_stmt;
	int ln;

	return_val_if_fail(node != NULL, NULL);

	for_node = (struct ast_node_for *)node;

	memset(&for_bool, 0, sizeof(struct bool_lists));
	memset(&for_stmt, 0, sizeof(struct stmt_lists));

	/* Explore the first optional expression. */
	if (for_node->optexpr1 != NULL)
		represent_node(for_node->optexpr1, NULL);
	/* Create a begin label. */
	ln = gen_label();
	label_begin = quad_label(ln);
	put_quad(QUAD(label_begin));
	/* Explore the second optional exprssion. */
	if (for_node->optexpr2 != NULL) {
		represent_node(for_node->optexpr2, &for_bool);
		ln = gen_label();
		label_true = quad_label(ln);
		put_quad(QUAD(label_true));
		/* Backpatch true expressions. */
		backpatch(for_bool.true_list, label_true->ln);
		slist_free_all(&for_bool.true_list, NULL);
	}
	/* The cycle body is always presented. */
	represent_node(for_node->stmt, &for_stmt);

	/* Create a next label and backpatch continue expressions. */
	if (for_stmt.continue_list != NULL) {
		ln = gen_label();
		label_next = quad_label(ln);
		put_quad(QUAD(label_next));
		backpatch(for_stmt.continue_list, label_next->ln);
		slist_free_all(&for_stmt.continue_list, NULL);
	}
	/* Explore the third optional expression. */
	if (for_node->optexpr3 != NULL)
		represent_node(for_node->optexpr3, NULL);
	/* Jump to the begin label. */
	ln = label_begin->ln;
	goto_begin = quad_goto(ln);
	put_quad(QUAD(goto_begin));

	label_false = NULL;
	/* We need to make 2 backpatches -- false expressions and break
	 * expressions. */
	if (for_node->optexpr2 != NULL && for_bool.false_list != NULL) {
		ln = gen_label();
		label_false = quad_label(ln);
		put_quad(QUAD(label_false));
		backpatch(for_bool.false_list, label_false->ln);
		slist_free_all(&for_bool.false_list, NULL);
	}

	if (for_stmt.break_list != NULL) {
		if (label_false == NULL) {
			ln = gen_label();
			label_false = quad_label(ln);
			put_quad(QUAD(label_false));
		}
		backpatch(for_stmt.break_list, label_false->ln);
		slist_free_all(&for_stmt.break_list, NULL);
	}

	return NULL;
}

/* A central routine, it calls nodes handlers. */
static struct quad *represent_node(struct ast_node *node, void *opaque)
{
	struct quad *q;
	unsigned int i;

	return_val_if_fail(node != NULL, NULL);

	for (i = 0; representers[i].node_type != NODE_TYPE_UNKNOWN; i++) {
		if (node->type == representers[i].node_type) {
			q = representers[i].handler(node, opaque);
			return q;
		}
	}

	if (node->child != NULL)
		represent_node(node->child, opaque);

	if (node->next != NULL)
		represent_node(node->next, opaque);

	return NULL;
}

/* Routine traverses the tree and converts it to the liner array. */
struct quad **tree2inter(struct ast_node_root *root, unsigned int *nquads)
{
	return_val_if_fail(root != NULL, NULL);

	represent_node(AST_NODE(root), NULL);
	*nquads = quads_used;

	return quads;
}

